  
"
    __________________________________________________
    @Alejandro F. Reimondo 13/03/98
    This proyect sources are property of : Alejandro F. Reimondo
    Its use is NOT legal without written rights.
    Amenabar 3476 Piso: 2 Dto: B.
    Ciudad de Buenos Aires.
    Argentina.
    Tel: (54)-1-703-5586.
    __________________________________________________
"
    (File pathName: 'LibTiff.Pol') fileIn;close.
!

PictureBitmap subclass: #TIFFImage
  instanceVariableNames: 
    ' mode reader '
  classVariableNames: ''
  poolDictionaries: 
    ' TIFFConstants '  !

Dynalink32 variableByteSubclass: #TIFFDLL32
  classVariableNames: ''
  poolDictionaries: 
    ' TIFFConstants ' !

ARObject subclass: #TIFFObject
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: ''  !

TIFFObject subclass: #TIFFExample
  instanceVariableNames: 
    ' notifier tiff '
  classVariableNames: ''
  poolDictionaries: ''  !

TIFFObject subclass: #TIFFMemoryReader
  instanceVariableNames: 
    ' stream imageSize readCallback writeCallback seekCallback closeCallback sizeCallback mapCallback unmapCallback '
  classVariableNames: ''
  poolDictionaries: '' !


!TIFFImage class methods !
   
allFilters
    "Private - Returns the filter specifications for all image files supported by the receiver.
    This method can be refined by subclasses..."

    ^ #(
        ('*.tif;*.tiff' 'TIFF Files (*.TIF)')
    )!

compressionCcittFax3
	"Returns value for this compression method."

	^CompressionCcittfax3!
  
compressionCcittFax4
	"Returns value for this compression method."

	^CompressionCcittfax4!
  
compressionCcittRle
	"Returns value for this compression method."

	^CompressionCcittrle!

compressionCcittRlew
	"Returns value for this compression method."

	^CompressionCcittrlew!
  
compressionJpeg
	"Returns value for this compression method."

	^CompressionJpeg!

compressionLzw
	"Returns value for this compression method."

	^CompressionLzw!
  
compressionNext
	"Returns value for this compression method."

	^CompressionNext!

compressionNone
	"Returns value for this compression method."

	^CompressionNone!

compressionPackBits
	"Returns value for this compression method."

	^CompressionPackbits!

compressionThunderscan
	"Returns value for this compression method."

	^CompressionThunderscan!
  
saveDIB: aDIB as: pathName compression: compression
	"Save aDIB as a TIF file named pathName with compression."

	| image result |
	image := self new.
	image pathName: pathName.
	result := image saveDIB: aDIB compression: compression.
	image release.
	^result!

test
    "Test the receiver displaying an image onto display.
    TIFFImage test
    "

    | image |
    image := self fromUser.
    image notNil ifTrue: [
        image asDIB inspect.
        image release.
    ].!
  
tiffLibrary
    " Private - Returns the TIFF library."

    ^TIFFDLL32 current! !



!TIFFImage methods !

asBitmap
    "Returns the receiver as a Bitmap."

    ^self asDIB asBitmap!
  
asDIB
    "Returns aDIB read from the receiver (or nil if failed)."

    |
    imageWidth imageLength bitsPerSample
    rowsPerStrip photometricInterpretation
    lineSize samplePerPixel align
    bitmapInfo bits library temp line
    byteStart byteStartJump nrow buffer dib
    |
    imageWidth := self width.
    imageWidth = 0 ifTrue: [ ^nil ].
    imageLength := self height.
    bitsPerSample := self bitsPerSample.
    rowsPerStrip := self rowsPerStrip.
    photometricInterpretation := self photometricInterpretation.
    lineSize := self scanlineSize.
    samplePerPixel := lineSize // imageWidth.

    align := 4 - (lineSize \\ 4). "Number of byte to add at the end of each line of the DIB"
    align == 4 ifTrue: [ align := 0 ].

    bitmapInfo := self
        newBitmapInfoWidth: imageWidth
        height: imageLength
        bitCount: self bitCount.

    bits := ByteArray new: bitmapInfo sizeImage.
    bits atAllPut: 255. "only for testing (white fill)..."
    dib := DIB new
        bitmapInfo: bitmapInfo;
        bits: bits;
        yourself.
    buffer := ByteArray new: self stripSize.
    library := self tiffLibrary.
    "starts at bottom and go to top..."
    byteStartJump := imageWidth * lineSize // imageWidth + align.
    byteStart := byteStartJump * (imageLength - 1) + 1.
    0 to: imageLength - 1 by: rowsPerStrip do: [:row|
        nrow := row + rowsPerStrip > imageLength
            ifTrue: [ imageLength - row ] ifFalse: [ rowsPerStrip ].
        (library
            readEncodedStrip: self asParameter
            number: (library computeStrip: self asParameter row: row sample: 0)
            into: buffer
            size: nrow * lineSize) = -1 ifTrue: [ ^nil ]. "read error"
        0 to: nrow - 1 do: [:rowIndex|
            line := rowIndex * lineSize + 1.
            samplePerPixel  = 3
            ifTrue: [
                0 to: imageWidth - 1 do: [:index|
                    temp := index * samplePerPixel.
                    bits at: byteStart + temp put: (buffer at: line + temp + 2).
                    bits at: byteStart + temp + 1 put: (buffer at: line + temp + 1).
                    bits at: byteStart + temp + 2 put: (buffer at: line + temp).
                ].
            ]
            ifFalse: [
                bits
                    replaceFrom: byteStart
                    to: byteStart + lineSize - 1
                    with: buffer
                    startingAt: line.
            ].
            byteStart := byteStart - byteStartJump "max: 1"
        ].
    ].

    ^dib!
 
bitCount
    "Returns the receiver's bitsCount"

    | bitCount |
    bitCount := self bitsPerSample * self scanlineSize // self width.

    "Make sure bits per pixel is valid"
    ^bitCount <= 1 ifTrue: [ 1 ]
        ifFalse: [ bitCount <= 4 ifTrue: [ 4 ]
        ifFalse: [ bitCount <= 8 ifTrue: [ 8 ]
        ifFalse: [ bitCount <= 24 ifTrue: [ 24 ]
        ifFalse: [ 4 ]]]
    ]!

bitsPerSample
    "Returns the receiver's bitsPerSample"

    ^self isValid
        ifTrue: [ self getIntegerField: TifftagBitspersample ]
        ifFalse: [0]!
   
boundingBox
    "Returns the receiver's boundingBox"

    ^0@0 extent: self extent!
  
colorMap
    " Private - Returns an array with structures pointing to color maps."

    | red green blue numberOfColors value grayScale |

    grayScale := [
        "color table is built for a gray scale homogeneous distribution"
        numberOfColors := self numberOfColors.
        numberOfColors = 2 ifTrue: [
            ^self mimimumValueIsBlack
            ifTrue: [ "no color table is needed. Standard B/W image." ]
            ifFalse: [ "reversed color r\table must be built."
                Array
                    with: (ByteArray with: 0 with: 255 with: 0 with: 0)
                    with: (ByteArray with: 0 with: 255 with: 0 with: 0)
                    with: (ByteArray with: 0 with: 255 with: 0 with: 0)
            ]
        ].
        red := WriteStream on: (ByteArray new: numberOfColors * 2).
        green := WriteStream on: (ByteArray new: numberOfColors * 2).
        blue := WriteStream on: (ByteArray new: numberOfColors * 2).
        0 to: numberOfColors - 1 do: [:i| "use 8 bit values"
            value := i + 1 * 255 // numberOfColors.
            red nextPut: 0;nextPut: value.
            green nextPut: 0;nextPut: value.
            blue nextPut: 0;nextPut: value.
        ].
        ^Array
            with: red contents
            with: green contents
            with: blue contents
    ].

    "PhotometricInterpretation = 1 image is gray scale."
    self photometricInterpretation = 1 ifTrue: [ ^grayScale value ].

    red := ExternalLong fromInteger: 0.
    green := ExternalLong fromInteger: 0.
    blue := ExternalLong fromInteger: 0.
    self tiffLibrary
        getField: self asParameter
        tag: TifftagColormap
        with: blue asParameter
        with: green asParameter
        with: red asParameter.

    (red asInteger = 0
    or: [ green asInteger = 0
    or: [ blue asInteger = 0 ]])
    ifTrue: [ ^grayScale value ].

    ^Array
        with: red asExternalAddress
        with: green asExternalAddress
        with: blue asExternalAddress!

colorMapOf: aDIB
	" Private - returns the color map of aDIB. "

	| stream red green blue |
	aDIB numberOfColors = 0 ifTrue: [ ^nil ].
	stream := ReadStream on: aDIB colorTable.
	red := WriteStream on: (ByteArray new: 256).
	green := WriteStream on: (ByteArray new: 256).
	blue := WriteStream on: (ByteArray new: 256).
	[stream atEnd] whileFalse: [
		blue nextPut:0;nextPut: stream next.
		green nextPut:0;nextPut: stream next.
		red nextPut:0;nextPut: stream next.
		stream next.
	].
	^Array
		with: red contents
		with: green contents
		with: blue contents
	!
 
colorTable
    " Private - Returns a byte array with color table compatible with Windows colorTable obtained from the receiver.
    Or nil if the receiver has no color table."

    ^self colorTableFor: self bitCount!

colorTableFor: bitCount
    " Private - Returns a byte array with color table compatible with Windows colorTable obtained from the receiver.
    Or nil if the receiver has no color table."

    | stream colorMap red green blue bytes paletteSize |
    self photometricInterpretation = PhotometricRgb ifTrue: [ ^nil ]. "image is RGB. No color table"
    "PhotometricInterpretation = PhotometricPalette  image have a color palette."
    "PhotometricInterpretation = 1 image is gray scale."
    bytes := ByteArray new: (WinBitmapInfo colorTableSize: bitCount).
    stream := WriteStream on: bytes.
    colorMap := self colorMap.
    colorMap isNil ifTrue: [ ^nil ].
    red := colorMap first.
    green := colorMap at: 2.
    blue := colorMap last.
    1 to: (1 bitShift: self bitsPerSample) do: [:index|
        stream
            nextPut: (red at: index * 2);
            nextPut: (green at: index * 2);
            nextPut: (blue at: index * 2);
            nextPut: 0;
            yourself
    ].
    ^bytes!
 
compression
    "Returns the receiver's compression"

    ^self isValid ifTrue: [ self getIntegerField: TifftagCompression ]!

copyToClipboard
    "Copies the receiver to the Clipboard"

    self isValid ifTrue: [
        ^self withBitmapDo: [:bitmap| Clipboard setBitmap: bitmap ]
    ]!
  
defaultStripSize: request
	"Returns the default strip size for the receiver's image."

	^self tiffLibrary
		defaultStripSize: self asParameter
		request: request asParameter!
 
dibFor: aRectangle
    "Private - Returns a DIB for the receiver's handle cropped to aRectangle...
    This method can be refined by subclasses..."

    ^self notImplementedYet!
   
displayAt: aPoint with: aPen
    "Display the receiver with aPen at aPoint."

    self isValid ifFalse: [^nil].
    ^self asDIB displayAt: aPoint with: aPen!
   
fileInFrom: aPathName format: aFormat
    "Private - Read the receiver internal structure from aFile.
    aFormat will be ignored."

    self privateHandle: (self tiffLibrary open: aPathName mode: self mode asParameter).
    self pathName: aPathName.!

flip: bits by: scanLineSize
    " Private - Flip the vertically the bits by scanLineSize."

    | buffer j |
    j := bits size.
    buffer := ByteArray new: scanLineSize.
    1 to: bits size - scanLineSize // 2 by: scanLineSize do: [:i|
        j := j - scanLineSize.
        buffer replaceFrom: 1 to: scanLineSize with: bits startingAt: i.
        bits
            replaceFrom: i to: i + scanLineSize - 1 with: bits startingAt: j + 1;
            replaceFrom: j + 1 to: j + scanLineSize with: buffer startingAt: 1
    ].
	^bits!
 
fromStream: aStream format: aFormat
    "Returns an instance of the receiver loaded from aStream or nil if failed..."

    | handle |
    reader := TIFFMemoryReader on: aStream.
    handle := reader open: (pathName isNil
        ifTrue: [ ('Image from ', aStream asString) ]
        ifFalse: [pathName]).
    self privateHandle: handle.!
   
getIntegerField: aTag
    " Private - Get the receiver's field with tag aTag. "

	^self tiffLibrary
        getIntegerField: self asParameter
        tag: aTag!
   
getIntegerFieldDefaulted: aTag
    " Private - Get the receiver's field with tag aTag. "

	^self tiffLibrary
        getIntegerFieldDefaulted: self asParameter
        tag: aTag!
 
getRGBA
    "Read the receiver and fill the aByteArray with RGBA contents."

    | buffer |
    buffer := ByteArray new: self rgbaSize.
    self getRGBAInto: buffer.
    ^buffer!

getRGBAInto: buffer
    "Read the receiver and fill the buffer (aByteArray or String) with RGBA contents."

    self tiffLibrary
        readRgbaImage: self asParameter
        width: self width
        height: self height
        raster: buffer
        stop: 0!
  
hasTiles
    "True if the receiver has tiles."

    ^self isValid and: [ self tiffLibrary isTiled: self asParameter ]!
   
height
    "Returns the receiver's height"

    ^self isValid
        ifTrue: [ self getIntegerField: TifftagImagelength ]
        ifFalse: [0]!
   
isHandleOk
    "True if the receiver has a valid handle"

    ^hImage notNil and: [hImage isValid]!
  
isTiled
    "True if the receiver has tiles."

    ^self hasTiles!
   
isUpSampled
    "True if the receiver is Up sampled."

    ^self isValid and: [ self tiffLibrary isUpSampled: self asParameter ]!

mimimumValueIsBlack
    "Returns True if the minimum value (0) corresponds to black color."

    ^self isValid and: [
        (self getIntegerField: TifftagPhotometric) = PhotometricMinisblack
    ]!

mode
    "Returns the instance var of the receiver
     for external manipulation"

    mode isNil ifTrue: [ self readWrite. ].
    ^mode!
 
mode: aMode
    "Sets the instance var of the receiver."

    mode := aMode.!

newBitmapInfoWidth: width height: height bitCount: bitCount
    " Private - Returns WinBitmapInfoHeader initialized to be part of a DIB of width@height compatible with the receiver."

    | bitmapInfo colorTable |

    bitmapInfo := WinBitmapInfo
        width: width height: height
        planes: 1 bitCount: bitCount.

   bitmapInfo
        compression: 0; "BI_RGB"
        sizeImage: bitmapInfo bitCount * width + 31 // 32 * 4 * height;
        xPelsPerMeter: 2834;
        yPelsPerMeter: 2834;
        clrUsed: 0;
        clrImportant: 0;
        yourself.

    (colorTable := self colorTable) notNil
    ifTrue: [ bitmapInfo colorTable: colorTable ]
    ifFalse: [
        bitmapInfo colorTable: #(0 0 0 0 255 255 255 0)
    ].

    ^bitmapInfo!
 
numberOfColors
    "Answer the number of colors in the receiver's color table."

    ^self bitCount = 24
        ifTrue: [ 0 ] "A 24 bit count bitmap has no color table."
        ifFalse: [ ^1 bitShift: self bitCount ]!

numberOfStrips
    "Returns the number of strips in the receiver's image."

    self isValid ifFalse: [ ^0 ].
    ^self tiffLibrary numberOfStrips: self asParameter!
   
numberOfTiles
    "Returns the number of tiles in the receiver's image."

    self isValid ifFalse: [ ^0 ].
    ^self tiffLibrary numberOfTiles: self asParameter!
  
pasteFromClipboard
    "Copies the Clipboard to the receiver"

    self fromBitmap: Clipboard getBitmap!
 
photometricInterpretation
    "Returns the receiver's photometricInterpretation"

    ^self isValid
        ifTrue: [ self getIntegerField: TifftagPhotometric ]
        ifFalse: [0]!
 
planarConfiguration
    "Returns the receiver's planar configuration."

    ^self getIntegerField: TifftagPlanarconfig!
  
privateHandle: aHandle
    "Private - Forces the receiver's handle"

    | old |
    aHandle isNumber ifTrue: [
        ^aHandle = 0
        ifTrue: [ self privateHandle: nil ]
        ifFalse: [ self privateHandle: (ExternalHandle fromInteger: aHandle) ]
    ].
    self isHandleOk ifTrue: [ old := hImage ].
    hImage := aHandle.
    old notNil ifTrue: [ self tiffLibrary close: old. ].!
   
readOnly
    "Sets the receiver for readOnly access."

    mode := 'rm'.!

readWrite
    "Sets the receiver for read/write access."

    mode := 'rwm'.!

release
    "Release the receiver's resources..."

    super release.
    self privateHandle: nil.
    reader release.
	reader := nil.!
   
reversedBitsOf: aDIB
    "Private - Revert the image bits of aDIB."

    ^self
		flip: aDIB bits copy
		by: (self scanLineSizeOf: aDIB)
!
 
rgbaSize
    "Returns the size needed for a byte array to hold RGBA contents."

    self isValid ifFalse: [ ^0 ].
    ^self width * self height * 4 "expanded to 32 bits"!
  
rowsPerStrip
    "Returns the receiver's rowsPerStrip"

    ^self isValid
        ifTrue: [ self getIntegerField: TifftagRowsperstrip ]
        ifFalse: [0]!
  
samplesPerPixel
    "Returns the number of samples per pixel."

    ^self planarConfiguration = PlanarconfigSeparate
        ifTrue: [ self getIntegerField: TifftagSamplesperpixel ]
        ifFalse: [ 0 ]!
  
save
    "Private - Save the receiver to current pathName."

	| dib result compression |
    dib := self asDIB.
	dib isNil ifTrue: [ ^false ].
	compression := self compression.
	self privateHandle: nil.
    result := self class
		saveDIB: dib
		as: self pathName
		compression: compression.
	self fileInFrom: self pathName format: nil.
	^result!
  
saveAs: aFileName
    "Private - Save the receiver to aFileName.
     Returns nil if failed..."

   ^self
        saveAs: aFileName
        format: nil!
  
saveAs: aFileName format: aFormat
    "Private - Save the receiver to aFileName."

	| dib |
    dib := self asDIB.
	dib isNil ifTrue: [ ^false ].
    ^self class
		saveDIB: dib
		as: aFileName
		compression: (aFormat isNil
			ifTrue: [ self compression ]
			ifFalse: [ aFormat ])!
 
saveDIB: aDIB compression: compression
	"Save aDIB into the receiver's pathName using compression method."

	^self saveDIB: aDIB compression: compression predictor: nil!

saveDIB: aDIB compression: compression predictor: predictor
    "Save aDIB into the receiver's pathName using compression method."

    | rowsPerStrip colorMap strip stripSize bits width extraBytes buffer samplesPerPixel swap |
    self writeOnly.
    self privateHandle: (self tiffLibrary
        open: self pathName
        mode: self mode asParameter).
	self isValid ifFalse: [
		^self error: 'Can''t open output file.'
	].
    width := aDIB width.
    self
        setIntegerField: TifftagImagewidth to: width;
        setIntegerField: TifftagImagelength to: aDIB height;
        setIntegerField: TifftagPhotometric to:
            (aDIB numberOfColors > 0
            ifTrue: [ samplesPerPixel := 1. PhotometricPalette ]
            ifFalse: [ samplesPerPixel := 3. PhotometricRgb ]);
        setIntegerField: TifftagPlanarconfig to: PlanarconfigContig;
        setIntegerField: TifftagSamplesperpixel to: samplesPerPixel;
        setIntegerField: TifftagBitspersample to: (aDIB bitCount // samplesPerPixel);
        yourself.
    rowsPerStrip := (self defaultStripSize: 0) min: aDIB height.
    self setIntegerField: TifftagRowsperstrip to: rowsPerStrip.
    compression notNil ifTrue: [
        self setIntegerField: TifftagCompression to: compression.
    ].
    (predictor notNil and: [predictor ~= 0])
    ifTrue: [ self setIntegerField: TifftagPredictor to: predictor ].
    colorMap := self colorMapOf: aDIB.
    colorMap notNil
    ifTrue: [
        self
            setField: TifftagColormap
            with: colorMap first
            with: (colorMap at: 2)
            with: colorMap last.
    ].
    self
        setIntegerField: TifftagOrientation to: OrientationTopleft;
        yourself.
    bits := ReadStream on: (self reversedBitsOf: aDIB).
    strip := 0.
    stripSize := self stripSize.
    buffer := ByteArray new: stripSize.
    extraBytes := (self scanLineSizeOf: aDIB) - (stripSize // rowsPerStrip).
    0 to: aDIB height - 1 by: rowsPerStrip do: [:row|
        extraBytes > 0
        ifTrue: [
            buffer := WriteStream on: buffer.
            rowsPerStrip timesRepeat: [
                buffer nextPutAll: (bits
                    next: (stripSize // rowsPerStrip
                        min: bits size - bits position)).
                bits atEnd ifFalse: [ bits next: extraBytes ].
            ].
            buffer := buffer contents.
        ]
        ifFalse: [
            stripSize := stripSize min: bits size - bits position.
            buffer := bits next: stripSize.
        ].
		aDIB bitCount = 24 ifTrue: [ "must reverse RGB values in buffer"
			1 to: buffer size -1 by: 3 do: [:i|
				swap := buffer at: i.
				buffer
					at: i put: (buffer at: i + 2);
					at: i + 2 put: swap;
					yourself
			].
		].
        self
            writeEncodedStrip: strip
            into: buffer size: buffer size.
        strip := strip + 1.
    ].
    "Invalidate receiver's use. (close the handle...)"
    self privateHandle: nil.!
   
scanlineSize
    "Returns the scan line size."

    self isValid ifFalse: [ ^0 ].
    ^self tiffLibrary scanlineSize: self asParameter!
 
scanLineSizeOf: aDIB
    "Private - Rturns the scan line size of aDIB."

    | bitmapInfo |
    bitmapInfo := aDIB bitmapInfo.
    ^bitmapInfo sizeImage = 0
    ifTrue: [ ((bitmapInfo width * bitmapInfo bitCount) + 31) // 32 * 4 ]
    ifFalse: [ bitmapInfo sizeImage // bitmapInfo height ]!
   
setField: aTag to: aValue
    " Private - Sets the receiver's field with tag aTag to aValue. "

    ^self tiffLibrary
        setField: self asParameter
        tag: aTag
        value: aValue asParameter!
 
setField: aTag with: aValue with: bValue with: cValue
    " Private - Sets the receiver's field with tag aTag to aValue,bValue,cValue. "

    ^self tiffLibrary
        setField: self asParameter
        tag: aTag
        with: aValue asParameter
		with: bValue asParameter
		with: cValue asParameter!

setIntegerField: aTag to: aValue
    " Private - Sets the receiver's field with tag aTag to aValue. "

    ^self tiffLibrary
        setIntegerField: self asParameter
        tag: aTag
        value: aValue asParameter!
   
stripSize
    "Returns the strip size."

    self isValid ifFalse: [ ^0 ].
    ^self tiffLibrary stripSize: self asParameter!
   
tiffLibrary
    " Private - Returns the TIFF library."

    ^self class tiffLibrary!
 
tileSize
    "Returns the tile size."

    self isValid ifFalse: [ ^0 ].
    ^self tiffLibrary tileSize: self asParameter!
  
width
    "Returns the receiver's width"

    ^self isValid
        ifTrue: [ self getIntegerField: TifftagImagewidth ]
        ifFalse: [0]!
  
withBitmapDo: aBlock
    "Private - Evaluates aBlock with a (ddb) bitmap obtainied from
     the receiver, then release the bitmap..."

    | bitmap result |
    self isValid ifFalse: [ ^nil ].
    bitmap := self asBitmap.
    result := aBlock value: bitmap.
    bitmap release.
    ^result!

withContentsDo: aBlock
    "Evaluates aBlock with a buffer for each streaming information in the receiver."

    self hasTiles
    ifTrue: [ self withTilesDo: aBlock ]
    ifFalse: [ self withStripsDo: aBlock ]
!
  
withDirectoriesDo: aBlock
    "Evaluates aBlock without parameters while traversing directories."

    self isValid ifFalse: [ ^nil ].
    [ self tiffLibrary readDirectory: self asParameter ]
    whileTrue: [ aBlock value ].
!

withScanLinesDo: aBlock
    "Evaluates aBlock with a scan line buffer for each scan line in the receiver."

    | buffer |
    buffer := ByteArray new: self scanlineSize.
    ^self withScanLinesIn: buffer do: aBlock
!
 
withScanLinesIn: buffer do: aBlock
    "Evaluates aBlock with:
        1.- a scan line buffer for each scan line in the receiver.
        2.- sample number (plane number)."

    | stop nSamples |
    self isValid ifFalse: [ ^nil ].
    stop := self height - 1.
    nSamples := self samplesPerPixel.
    0 to: nSamples do: [:sample|
        0 to: stop do: [:row|
            self tiffLibrary
                readScanLine: self asParameter
                into: buffer
                row: row
                mode: sample.
            aBlock value: buffer value: sample
        ]
    ]
!
   
withStripsDo: aBlock
    "Evaluates aBlock with a strip buffer for each strip in the receiver."

    | buffer |
    buffer := ByteArray new: self stripSize.
    ^self withStripsIn: buffer do: aBlock
!
  
withStripsIn: buffer do: aBlock
    "Evaluates aBlock with a strip buffer for each strip in the receiver."

    | stop |
    self isValid ifFalse: [ ^nil ].
    stop := self numberOfStrips - 1.
    0 to: stop do: [:strip|
        self tiffLibrary
            readEncodedStrip: self asParameter
            number: strip asParameter
            into: buffer
            size: -1 asParameter.
        aBlock value: buffer
    ]
!
   
withTilesDo: aBlock
    "Evaluates aBlock with a strip buffer for each tile in the receiver."

    | buffer |
    buffer := ByteArray new: self tileSize.
    ^self withTilesIn: buffer do: aBlock
!
  
withTilesIn: buffer do: aBlock
    "Evaluates aBlock with a strip buffer for each tile in the receiver."

    | stop |
    self isValid ifFalse: [ ^nil ].
    stop := self numberOfTiles - 1.
    0 to: stop do: [:tile|
        self tiffLibrary
            readEncodedTile: self asParameter
            number: tile asParameter
            into: buffer
            size: -1 asParameter.
        aBlock value: buffer
    ]
!
 
writeEncodedStrip: aNumber into: lpBuffer size: size
	"Encode suplied data and write into the specified strip."

    ^self tiffLibrary
		writeEncodedStrip: self asParameter
		strip: aNumber
		into: lpBuffer
		size: size!
 
writeOnly
    "Sets the receiver for write access."

    mode := 'wm'.! !



!TIFFDLL32 class methods !
  
callbackTable
    " Private - Returns the callback shape table for defining callback handlers."

    ^#(
    "(callBackName (arguments...) returnType callingConvertion)"

    (errorHandler (ulong ulong ulong) none c) "typedef    void (*TIFFErrorHandler)(const char*, const char*, va_list);"
    (readWriteProc (long long handle) long c) "typedef    int32 (*TIFFReadWriteProc)(handle, lpData, int32);"
    (seekProc (long long handle) long c) "typedef    int32 (*TIFFSeekProc)(handle, int32, int);"
    (closeProc (handle) long c) "typedef    int (*TIFFCloseProc)(handle);"
    (sizeProc (handle) long c) "typedef    int32 (*TIFFSizeProc)(handle);"
    (mapFileProc (ulong ulong handle) long c) "typedef    int (*TIFFMapFileProc)(handle, lpData*, int32*);"
    (unmapFileProc (long ulong handle) none c) "typedef    void (*TIFFUnmapFileProc)(handle, lpData, int32);"
    (extendProc (ulong) none c) "typedef    void (*TIFFExtendProc)(TIFF*);"



    "    The image reading and conversion routines invoke put routines to copy/image/whatever tiles of raw image data.
    A default set of routines are provided to convert/copy raw image data to 8-bit packed ABGR format rasters.
    Applications can supply alternate routines that unpack the data into a different format or, for example, unpack the data nd draw the unpacked raster on the display.
    See TIFFRGBAImage structure.
    typedef void (*tileContigRoutine)            (TIFFRGBAImage*, uint32*, uint32, uint32, uint32, uint32, int32, int32, unsigned char*);
    typedef void (*tileSeparateRoutine)    (TIFFRGBAImage*, uint32*, uint32, uint32, uint32, uint32, int32, int32, unsigned char*, unsigned char*, unsigned char*, unsigned char*);"
    (putContigRGBAImage (ulong ulong ulong ulong ulong ulong ulong ulong ulong) none c)
    (putSeparateRGBAImage (ulong ulong ulong ulong ulong ulong ulong ulong ulong ulong ulong ulong) none c)
    (putAnyRGBAImage (ulong) none c)
    (getRGBAImage (ulong ulong ulong ulong) long c)

    "A CODEC is a software package that implements decoding, encoding, or decoding+encoding of a compression algorithm.
    The library provides a collection of builtin codecs.
    More codecs may be registered through calls to the library and/or the builtin implementations may be overridden."
    (codecInitMethod (long ulong) long c)

    )!
 
fileName
    "Private - Answer the receiver's file name."

"
	Terminal bell.
	(MessageBox confirm: 'Use debug version of LibTiff.dll ?')
	ifTrue: [ ^'\ostyle\tiff\libTiff\libTiff_\LibTiff.dll' ].
"	
     ^'LibTiff.Dll'!
 
initializeConstants
   "Private - Initializes the constants for the interface...
        TIFFDLL32 initializeConstants.
    "

    TIFFConstants size = 0 ifTrue: [
        TIFFConstants := Dictionary new
    ].

  TIFFConstants
     at: 'TifftagDocumentname' put: 269;
    at: 'PhotometricPalette' put: 3;
    at: 'PlanarconfigSeparate' put: 2;
    at: 'OrientationRightbot' put: 7;
    at: 'TifftagRefpts' put: 32953;
    at: 'TifftagYcbcrpositioning' put: 531;
    at: 'TifftagOrientation' put: 274;
    at: 'SampleformatVoid' put: 4;
    at: 'OrientationTopleft' put: 1;
    at: 'TifftagJpegqtables' put: 519;
    at: 'TifftagSampleformat' put: 339;
    at: 'TifftagDatetime' put: 306;
    at: 'TifftagImagewidth' put: 256;
    at: 'PhotometricMask' put: 4;
    at: 'PhotometricRgb' put: 2;
    at: 'TiffBigendian' put: 19789;
    at: 'Grayresponseunit10s' put: 1;
    at: 'ExtrasampleUnassalpha' put: 2;
    at: 'Grayresponseunit100000s' put: 5;
    at: 'CompressionCcittrlew' put: 32771;
    at: 'TifftagXposition' put: 286;
    at: 'PhotometricMinisblack' put: 1;
    at: 'OrientationBotleft' put: 4;
    at: 'ThreshholdBilevel' put: 1;
    at: 'TifftagExtrasamples' put: 338;
    at: 'TifftagJpegrestartinterval' put: 515;
    at: 'TifftagSminsamplevalue' put: 340;
    at: 'JpegprocLossless' put: 14;
    at: 'FiletypeReducedimage' put: 1;
    at: 'PlanarconfigContig' put: 1;
    at: 'TifftagGrayresponsecurve' put: 291;
    at: 'Colorresponseunit10s' put: 1;
    at: 'TifftagYcbcrcoefficients' put: 529;
    at: 'PhotometricYcbcr' put: 6;
    at: 'TifftagGroup3options' put: 292;
    at: 'TifftagPrimarychromaticities' put: 319;
    at: 'ExtrasampleAssocalpha' put: 1;
    at: 'Colorresponseunit100000s' put: 5;
    at: 'TifftagJpegpointtransform' put: 518;
    at: 'YcbcrpositionCentered' put: 1;
    at: 'TifftagImagedepth' put: 32997;
    at: 'TifftagMake' put: 271;
    at: 'CompressionCcittfax3' put: 3;
    at: 'TifftagBitspersample' put: 258;
    at: 'Group3opt2dencoding' put: 1;
    at: 'CleanfaxdataRegenerated' put: 1;
    at: 'ExtrasampleUnspecified' put: 0;
    at: 'ResunitInch' put: 2;
    at: 'Group4optUncompressed' put: 2;
    at: 'TifftagJpegproc' put: 512;
    at: 'Group3optFillbits' put: 4;
    at: 'TifftagCleanfaxdata' put: 327;
    at: 'TifftagOsubfiletype' put: 255;
    at: 'Tifftagresolution' put: 283;
    at: 'FiletypePage' put: 2;
    at: 'TifftagConsecutivebadfaxlines' put: 328;
    at: 'Colorresponseunit10000s' put: 4;
    at: 'TifftagSubfiletype' put: 254;
    at: 'TifftagTilewidth' put: 322;
    at: 'TifftagSoftware' put: 305;
    at: 'FillorderLsb2msb' put: 2;
    at: 'CompressionCcittrle' put: 2;
    at: 'TifftagFreeoffsets' put: 288;
    at: 'InksetCmyk' put: 1;
    at: 'TifftagInknames' put: 333;
    at: 'TiffVersion' put: 42;
    at: 'TifftagTilebytecounts' put: 325;
    at: 'Grayresponseunit100s' put: 2;
    at: 'CompressionThunderscan' put: 32809;
    at: 'TifftagGrayresponseunit' put: 290;
    at: 'PhotometricSeparated' put: 5;
    at: 'SampleformatInt' put: 2;
    at: 'TifftagTilelength' put: 323;
    at: 'TifftagMaxsamplevalue' put: 281;
    at: 'TifftagFreebytecounts' put: 289;
    at: 'Colorresponseunit1000s' put: 3;
    at: 'OfiletypeReducedimage' put: 2;
    at: 'TifftagRegiontackpoint' put: 32954;
    at: 'OrientationLefttop' put: 5;
    at: 'TifftagPlanarconfig' put: 284;
    at: 'OrientationRighttop' put: 6;
    at: 'TifftagTargetprinter' put: 337;
    at: 'TifftagCompression' put: 259;
    at: 'TifftagCelllength' put: 265;
    at: 'ThreshholdHalftone' put: 2;
    at: 'TifftagRegionaffine' put: 32956;
    at: 'CompressionPackbits' put: 32773;
    at: 'Grayresponseunit1000s' put: 3;
    at: 'CleanfaxdataUnclean' put: 2;
    at: 'OrientationTopright' put: 2;
    at: 'TifftagReferenceblackwhite' put: 532;
    at: 'TifftagYposition' put: 287;
    at: 'TifftagMatteing' put: 32995;
    at: 'CompressionLzw' put: 5;
    at: 'TifftagColormap' put: 320;
    at: 'TiffLittleendian' put: 18761;
    at: 'OfiletypeImage' put: 1;
    at: 'TifftagImagelength' put: 257;
    at: 'TifftagPagename' put: 285;
    at: 'TifftagGroup4options' put: 293;
    at: 'OrientationBotright' put: 3;
    at: 'TifftagTransferfunction' put: 301;
    at: 'TifftagColorresponseunit' put: 300;
    at: 'Grayresponseunit10000s' put: 4;
    at: 'CompressionJpeg' put:7;
    at: 'TifftagModel' put: 272;
    at: 'TifftagDatatype' put: 32996;
    at: 'TifftagResolutionunit' put: 296;
    at: 'TifftagRowsperstrip' put: 278;
    at: 'TifftagStripoffsets' put: 273;
    at: 'PhotometricCielab' put: 8;
    at: 'TifftagSamplesperpixel' put: 277;
    at: 'TifftagWhitepoint' put: 318;
    at: 'SampleformatUint' put: 1;
    at: 'YcbcrpositionCosited' put: 2;
    at: 'ResunitNone' put: 1;
    at: 'CompressionNone' put: 1;
    at: 'TifftagHostcomputer' put: 316;
    at: 'TifftagMinsamplevalue' put: 280;
    at: 'TifftagPredictor' put: 317;
    at: 'TifftagYcbcrsubsampling' put: 530;
    at: 'OfiletypePage' put: 3;
    at: 'TifftagDotrange' put: 336;
    at: 'Colorresponseunit100s' put: 2;
    at: 'TifftagJpegifbytecount' put: 514;
    at: 'TifftagTileoffsets' put: 324;
    at: 'TifftagJpegactables' put: 521;
    at: 'TifftagXresolution' put: 282;
    at: 'PhotometricMiniswhite' put: 0;
    at: 'SampleformatIeeefp' put: 3;
    at: 'TifftagJpeglosslesspredictors' put: 517;
    at: 'ResunitCentimeter' put: 3;
    at: 'CompressionNext' put: 32766;
    at: 'TifftagInkset' put: 332;
    at: 'FiletypeMask' put: 4;
    at: 'FillorderMsb2lsb' put: 1;
    at: 'TifftagPagenumber' put: 297;
    at: 'TifftagRegionwarpcorners' put: 32955;
    at: 'ThreshholdErrordiffuse' put: 3;
    at: 'TifftagTiledepth' put: 32998;
    at: 'TifftagPhotometric' put: 262;
    at: 'TifftagJpegifoffset' put: 513;
    at: 'TifftagFillorder' put: 266;
    at: 'TifftagArtist' put: 315;
    at: 'TifftagCellwidth' put: 264;
    at: 'CleanfaxdataClean' put: 0;
    at: 'OrientationLeftbot' put: 8;
    at: 'JpegprocBaseline' put: 1;
    at: 'TifftagBadfaxlines' put: 326;
    at: 'TifftagSmaxsamplevalue' put: 341;
    at: 'CompressionCcittfax4' put: 4;
    at: 'TifftagHalftonehints' put: 321;
    at: 'TifftagImagedescription' put: 270;
    at: 'Group3optUncompressed' put: 2;
    at: 'TifftagJpegdctables' put: 520;
    at: 'TifftagThreshholding' put: 263;
    at: 'TiffprintStrips' put: 1;
    at: 'TiffprintCurves' put: 2;
    at: 'TiffprintJpegactables' put: 512;
    at: 'TiffprintJpegdctables' put: 512;
    at: 'TiffprintColormap' put: 4;
    at: 'TiffprintNone' put: 0;
    at: 'TiffprintJpegqtables' put: 256;
    at: 'FieldStripbytecounts' put: 26;
    at: 'FieldMake' put: 13;
    at: 'FieldImagedepth' put: 41;
    at: 'FieldTiledimensions' put: 1;
    at: 'FieldSampleformat' put: 38;
    at: 'TiffDirtydirect' put: 2;
    at: 'TiffOpt7' put: 128;
    at: 'FieldBadfaxrun' put: 37;
    at: 'TiffBuffersetup' put: 4;
    at: 'FieldResolutionunit' put: 24;
    at: 'TiffBeenwriting' put: 8;
    at: 'FieldPrimarychromas' put: 54;
    at: 'FieldJpegproc' put: 47;
    at: 'FieldGroup4options' put: 23;
    at: 'FieldBadfaxlines' put: 35;
    at: 'FieldTiledepth' put: 42;
    at: 'TiffIstiled' put: 128;
    at: 'FieldTargetprinter' put: 59;
    at: 'TiffDirtyheader' put: 1;
    at: 'FieldWhitepoint' put: 53;
    at: 'FieldPhotometric' put: 8;
    at: 'FieldPagename' put: 21;
    at: 'FieldYcbcrcoefficients' put: 44;
    at: 'FieldDatetime' put: 31;
    at: 'TiffNobitrev' put: 32;
    at: 'TiffOpt0' put: 1;
    at: 'FieldInknames' put: 57;
    at: 'TiffOpt1' put: 2;
    at: 'FieldPosition' put: 4;
    at: 'FieldMaxsamplevalue' put: 19;
    at: 'TiffOpt2' put: 4;
    at: 'TiffSwab' put: 16;
    at: 'FieldGroup3options' put: 22;
    at: 'TiffOpt3' put: 8;
    at: 'FieldStripoffsets' put: 27;
    at: 'FieldJpegqtables' put: 49;
    at: 'FieldSmaxsamplevalue' put: 40;
    at: 'TiffOpt4' put: 16;
    at: 'FieldDocumentname' put: 11;
    at: 'FieldThreshholding' put: 9;
    at: 'FieldCompression' put: 7;
    at: 'TiffOpt5' put: 32;
    at: 'FieldLast' put: 59;
    at: 'FieldImagedescription' put: 12;
    at: 'FieldSubfiletype' put: 5;
    at: 'FieldJpegdctables' put: 50;
    at: 'TiffOpt6' put: 64;
    at: 'FieldPredictor' put: 29;
    at: 'FieldJpegrestartinterval' put: 48;
    at: 'FieldPlanarconfig' put: 20;
    at: 'FieldHostcomputer' put: 32;
    at: 'FieldModel' put: 14;
    at: 'FieldOrientation' put: 15;
    at: 'FieldRefblackwhite' put: 52;
    at: 'FieldColormap' put: 28;
    at: 'TiffMybuffer' put: 64;
    at: 'FieldTransferfunction' put: 55;
    at: 'FieldHalftonehints' put: 43;
    at: 'FieldArtist' put: 30;
    at: 'FieldYcbcrpositioning' put: 46;
    at: 'FieldJpegactables' put: 51;
    at: 'TiffPostencode' put: 512;
    at: 'FieldBitspersample' put: 6;
    at: 'FieldSamplesperpixel' put: 16;
    at: 'FieldYcbcrsubsampling' put: 45;
    at: 'FieldSminsamplevalue' put: 39;
    at: 'FieldSoftware' put: 33;
    at: 'FieldCelldimensions' put: 2;
    at: 'TiffMapped' put: 256;
    at: 'FieldCleanfaxdata' put: 36;
    at: 'FieldMinsamplevalue' put: 18;
    at: 'FieldPagenumber' put: 25;
    at: 'FieldRowsperstrip' put: 17;
    at: 'FieldResolution' put: 3;
    at: 'FieldFillorder' put: 10;
    at: 'FieldImagedimensions' put: 0;
    at: 'FieldExtrasamples' put: 34;
    at: 'FieldInkset' put: 56;
    at: 'FieldDotrange' put: 58;
    yourself.
!
   
initializeStructures
    "Private - Initializes the self defined structures for the DLL.
    TIFFDLL32 initializeStructures.
    "

    self
        registerType: 'TIFFRGBAImage'
        as: #(
            (tiff handle) "tiff image handle"
            (stoponerr long) "stop on read error"
            (isContig long) "data is packed/separate"
            (alpha long) "type of alpha data present"
            (width ulong) "image width"
            (height ulong) "image height"
            (bitspersample ulong) "image bits/sample"
            (samplesperpixel ulong) "image samples/pixel"
            (orientation ulong) "image orientation"
            (photometric ulong) "image photometric interp"
            (redcmap long asWinAddress) "colormap pallete"
            (greencmap long asWinAddress)
            (bluecmap long asWinAddress)
            (get ulong asWinAddress) "get image data routine"
            (putAny ulong asWinAddress) "put decoded strip/tile routines"
            (putContig ulong asWinAddress)
            (putSeparate ulong asWinAddress)
            (map long asWinAddress) "sample mapping array"
            (bwMap long asWinAddress) "black&white map"
            (palMap long asWinAddress) "palette image map"
            (ycbcr '*TIFFYCbCrToRGB') "YCbCr conversion state"
            );
        registerType: 'TIFFYCbCrToRGB' "YCbCr->RGB support"
        as: #(
            (clamptab long asWinAddress) "range clamping table"
            (CrRTab (0 long))
            (CbBTab (0 long))
            (CrGTab (0 long))
            (CbGTab (0 long))
            (coeffs 18 asString) "float    coeffs[3]; cached for repeated use"
            );

    "A CODEC is a software package that implements decoding, encoding, or decoding+encoding of a compression algorithm.
    The library provides a collection of builtin codecs.
    More codecs may be registered through calls to the library and/or the builtin implementations may be overridden."
        registerType: 'TIFFCodec'
        as: #(
            (name long asWinAddress)
            (scheme ushort)
            (initMethod long asWinAddress)
            );
        yourself!

newCallBack: callBackName sending: selector to: aReceiver
    "Returns a CallBack object for sending message with selector to aReceiver.
    callBackName must be a valid name defined in callbackTable."

    | tuple aMessage parameterTypes |
    tuple := self callbackTable
        detect: [:one| one first = callBackName ]
        ifNone: [ ^self error: 'Invalid callback name.'].

    "check selector arguments with parameterTypes size"
    parameterTypes := tuple at: 2.
    (selector asString select: [:c| c = $: ]) size = parameterTypes size
    ifFalse: [ ^self error: 'Invalid selector (Must have ',parameterTypes size asString,' arguments.' ].

    aMessage := Message new
        receiver: aReceiver;
        selector: selector;
        yourself.
    ^CallBack
        registerMessage: aMessage
        parameterTypes: parameterTypes
        returnType: (tuple at: 3)
        callingConvention: (tuple size > 3 ifTrue: [ tuple at: 4 ] ifFalse: [#api]).
! !



!TIFFDLL32 methods !
 
checkTile: tifHandle x: x y: y z: z sample: sample
    "Check an (x,y,z,s) coordinate against the image bounds.
    ('sample' is a sample number)"

    <api: TIFFCheckTile handle ulong ulong ulong ushort boolean>
    ^self invalidArgument!

clientOpen: fileName mode: mode
    clientdata: handle
    readproc: readproc writeproc: writeproc seekproc: seekproc closeproc: closeproc
    sizeproc: sizeproc mapproc: mapproc unmapproc: unmapproc
    "Open a file for reading or writing.
    Returns a Tiff handle or zero if failed."

    | result |
    result := self
        tiffClientOpen: fileName mode: mode
        clientdata: handle
        readproc: readproc writeproc: writeproc
        seekproc: seekproc closeproc: closeproc
        sizeproc: sizeproc
        mapproc: mapproc unmapproc: unmapproc.
    ^result = 0 ifFalse: [ ExternalHandle fromInteger: result ]!
  
close: tiffHandle
    "Close an open Tiff element."

    <api: TIFFClose handle none>
    ^self invalidArgument!

codecAtAddress: anAddress
    " Private - Returns the codec structure at address anAddress (or nil)."

    (anAddress isNil or: [anAddress = 0 ]) ifTrue: [ ^nil ].
    anAddress isExternalAddress
    ifFalse: [ ^self codecAtAddress: anAddress asExternalAddress ].
    ^self class
        structureNamed: 'TIFFCodec'
        atAddress: anAddress!
   
computeStrip: tiffHandle row: row sample: sample
    "Compute which strip a (row,sample) value is in."

    <api: TIFFComputeStrip handle ulong ushort ulongReturn>
    ^self invalidArgument!
  
computeTile: tiffHandle x: x y: y z: z sample: sample
    "Compute which tile an (x,y,z,s) value is in.."

    <api: TIFFComputeTile handle ulong ulong ulong ushort ulong>
    ^self invalidArgument!
  
currentDirectory: tifHandle
    "Return index of the current directory."

    <api: TIFFCurrentDirectory handle ushort>
    ^self invalidArgument!
  
currentDirOffset: tifHandle
    "Return file offset of the current directory."

    <api: TIFFCurrentDirOffset handle ulongReturn>
    ^self invalidArgument!
   
currentRow: tifHandle
    "Return current row being read/written."

    <api: TIFFCurrentRow handle ulongReturn>
    ^self invalidArgument!
 
currentStrip: tifHandle
    "Returns current strip number (0 relative)"

    <api: TIFFCurrentStrip handle ulongReturn>
    ^self invalidArgument!
  
currentTile: tifHandle
    "Return current tile number (0 relative)."

    <api: TIFFCurrentTile handle ulongReturn>
    ^self invalidArgument!
 
defaultDirectory: tifHandle
    "Setup a default directory structure."

    <api: TIFFDefaultDirectory handle ulong>
    ^self invalidArgument!
 
defaultStripSize: tifHandle request: request
    "Compute a default strip size based on the image characteristics and a requested value.
    If the request is <1 then we choose a strip size according to certain heuristics."

    <api: TIFFDefaultStripSize handle ulong ulongReturn>
    ^self invalidArgument!
   
defaultTileSize: tifHandle width: width height: height
    " Private - Compute a default tile size based on the image characteristics and a requested value.
    If the request is <1 then we choose a strip size according to certain heuristics."

    <api: TIFFDefaultTileSize handle struct struct none>
    ^self invalidArgument!
   
fileDescriptor: aFileHandle open: fileName mode: mode
    " Private - Open a TIFF on an opened file descriptor for read/writing.
    Note that TIFFFdOpen and TIFFOpen recognise the character 'u' in the mode
    string, which forces the file to be opened unmapped."

    <api: TIFFFdOpen ulong struct struct long>
    ^self invalidArgument!
   
fileName: tifHandle
    "Return open file's name."

    <api: TIFFFileName handle long>
    ^self invalidArgument!
  
fileno: tifHandle
    "Return open file's I/O descriptor."

    <api: TIFFFileno handle long>
    ^self invalidArgument!

findCodec: scheme
    " Private - Returns the codec pointer for scheme codec."

    <api: TIFFFindCODEC ushort ulongReturn>
    ^self invalidArgument!
  
flush: tiffHandle
    "Flush any buffered information to file."

    <api: TIFFFlush handle boolean>
    ^self invalidArgument!
 
flushData: tiffHandle
    "Flush buffered data to the file."

    <api: TIFFFlushData handle boolean>
    ^self invalidArgument!

flushData1: tiffHandle
    "Flush buffered data to the file.
    Internal version of TIFFFlushData that can be called by 'encodestrip routines' w/o concern for infinite recursion."

    <api: TIFFFlushData1 handle boolean>
    ^self invalidArgument!
  
freeDirectory: tiffHandle
    "Release storage associated with a directory."

    <api: TIFFFreeDirectory handle none>
    ^self invalidArgument!
   
getAttribute: index
    "Extract attribute component from the packed ABGR form returned by TIFFReadRGBAImage."

    ^(index bitShift: -24) bitAnd: 16rFF!

getBitRevTable: reversed
    "Bit reversal tables.
    TIFFBitRevTable[<byte>] gives the bit reversed value of <byte>.
    Used in various places in the library when the FillOrder requires bit reversal of byte values (e.g. CCITT Fax 3 encoding/decoding).
    TIFFNoBitRevTable is provided for algorithms that want an equivalent table that do not reverse bit values.
    Returns anExternalAddress pointing to TIFFBitRevTable array or nil."

    | result |
    result := self tiffGetBitRevTable: reversed.
    ^result = 0 ifFalse: [ result asExternalAddress ]!
 
getBlue: index
    "Extract blue component from the packed ABGR form returned by TIFFReadRGBAImage."

    ^(index bitShift: -16) bitAnd: 16rFF!
  
getCodec: scheme
    "Returns the codec structure for scheme codec (or nil)."

    ^self codecAtAddress: (self findCodec: scheme)!
   
getDefaultTileSize: tifHandle
    "Compute a default tile size based on the image characteristics and a requested value.
    If the request is <1 then we choose a strip size according to certain heuristics."

    | width height |
    width := ExternalLong fromInteger: 0.
    height := ExternalLong fromInteger: 0.
    self
        defaultTileSize: tifHandle
        width: width asParameter
        height: height asParameter.
    ^width asInteger @ height asInteger!
 
getField: tiffHandle tag: tag value: value
    " Private - Returns a tag's value in the current tiff directory. "

    <api: TIFFGetField handle long struct boolean>
    ^self invalidArgument!

getField: tiffHandle tag: tag with: value1 with: value2 with: value3
    " Private - Returns a tag's value in the current tiff directory. "

    <api: TIFFGetField handle long struct struct struct boolean>
    ^self invalidArgument!

getFieldDefaulted: tiffHandle tag: tag value: value
    " Private - Like TIFFGetField, but return any default value if the tag is not present in the directory."

    <api: TIFFGetFieldDefaulted handle long struct boolean>
    ^self invalidArgument!

getGreen: index
    "Extract green component from the packed ABGR form returned by TIFFReadRGBAImage."

    ^(index bitShift: -8) bitAnd: 16rFF!
 
getIntegerField: tiffHandle tag: tag
    "Returns a tag's value in the current tiff directory."

    | long |
    long := ExternalLong fromInteger: 0.
    self getField: tiffHandle tag: tag value: long asParameter.
    ^long asInteger!
   
getIntegerFieldDefaulted: tiffHandle tag: tag
    "Returns a tag's value in the current tiff directory returning a default value if not present."

    | long |
    long := ExternalLong fromInteger: 0.
    self getFieldDefaulted: tiffHandle tag: tag value: long asParameter.
    ^long asInteger!

getMode: tifHandle
    "Return read/write mode."

    <api: TIFFGetMode handle long>
    ^self invalidArgument!
 
getRed: index
    "Extract red component from the packed ABGR form returned by TIFFReadRGBAImage."

    ^index bitAnd: 16rFF!

getVersion
    " Private - Get the version constant string address pointer."

    <api: TIFFGetVersion long>
    ^self invalidArgument!
 
initCCITTFax3: tifHandle scheme: scheme
    "..."

    <api: TIFFInitCCITTFax3 handle long boolean>
    ^self invalidArgument!
  
initCCITTFax4: tifHandle scheme: scheme
    "..."

    <api: TIFFInitCCITTFax4 handle long boolean>
    ^self invalidArgument!
  
initCCITTRLE: tifHandle scheme: scheme
    "..."

    <api: TIFFInitCCITTRLE handle long boolean>
    ^self invalidArgument!

initCCITTRLEW: tifHandle scheme: scheme
    "..."

    <api: TIFFInitCCITTRLEW handle long boolean>
    ^self invalidArgument!
  
initDumpMode: tifHandle scheme: scheme
    "Initialize dump mode."

    <api: TIFFInitDumpMode handle long boolean>
    ^self invalidArgument!
  
initJPEG: tifHandle scheme: scheme
    "..."

    <api: TIFFInitJPEG handle long boolean>
    ^self invalidArgument!

initLZW: tifHandle scheme: scheme
    "..."

    <api: TIFFInitLZW handle long boolean>
    ^self invalidArgument!
  
initPackBits: tifHandle scheme: scheme
    "..."

    <api: TIFFInitPackBits handle long boolean>
    ^self invalidArgument!

initThunderScan: tifHandle scheme: scheme
    "..."

    <api: TIFFInitThunderScan handle long boolean>
    ^self invalidArgument!
  
isByteSwapped: tifHandle
    "Return true if the data is returned byte swapped."

    <api: TIFFIsByteSwapped handle boolean>
    ^self invalidArgument!

isMSB2LSB: tifHandle
    "Return true if the data is returned in MSB-to-LSB bit order."

    <api: TIFFIsMSB2LSB handle boolean>
    ^self invalidArgument!
 
isTiled: tifHandle
    "Return true if file is organized in tiles. (False if organized as strips.)"

    <api: TIFFIsTiled handle boolean>
    ^self invalidArgument!
   
isUpSampled: tifHandle
    "Return true if the data is returned up-sampled."

    <api: TIFFIsUpSampled handle boolean>
    ^self invalidArgument!
  
lastDirectory: tifHandle
    "Return an indication of whether or not we are at the last directory in the file."

    <api: TIFFLastDirectory handle boolean>
    ^self invalidArgument!
 
numberOfStrips: tifHandle
    "Compute how many strips are in an image."

    <api: TIFFNumberOfStrips handle ulongReturn>
    ^self invalidArgument!
   
numberOfTiles: tifHandle
    "Compute how many tiles are in an image."

    <api: TIFFNumberOfTiles handle ulongReturn>
    ^self invalidArgument!
  
open: fileName mode: mode
    "Open a TIFF file for reading/writing.

    The following flags may be used to control intrinsic library
    behaviour that may or may not be desirable (usually for
    compatibility with some application that claims to support
    TIFF but only supports some braindead idea of what the
    vendor thinks TIFF is):

    l    use little-endian byte order for creating a file
    b    use big-endian byte order for creating a file
    L    read/write information using LSB2MSB bit order
    B    read/write information using MSB2LSB bit order
    H    read/write information using host bit order
    M    enable use of memory-mapped files when supported
    m    disable use of memory-mapped files
    C    enable strip chopping support when reading
    c    disable strip chopping support

    The use of the 'l' and 'b' flags is strongly discouraged.
    These flags are provided solely because numerous vendors,
    typically on the PC, do not correctly support TIFF; they
    only support the Intel little-endian byte order.  This
    support is not configured by default because it supports
    the violation of the TIFF spec that says that readers *MUST*
    support both byte orders.  It is strongly recommended that
    you not use this feature except to deal with busted apps
    that write invalid TIFF.  And even in those cases you should
    bang on the vendors to fix their software.

    The 'L', 'B', and 'H' flags are intended for applications
    that can optimize operations on data by using a particular
    bit order.  By default the library returns data in MSB2LSB
    bit order for compatibiltiy with older versions of this
    library.  Returning data in the bit order of the native cpu
    makes the most sense but also requires applications to check
    the value of the FillOrder tag; something they probabyl do
    not do right now.

    The 'M' and 'm' flags are provided because some virtual memory
    systems exhibit poor behaviour when large images are mapped.
    These options permit clients to control the use of memory-mapped
    files on a per-file basis.

    The 'C' and 'c' flags are provided because the library support
    for chopping up large strips into multiple smaller strips is not
    application-transparent and as such can cause problems.  The 'c'
    option permits applications that only want to look at the tags,
    for example, to get the unadulterated TIFF tag information.

    Returns a handle or nil if failed."

    | result |
    result := self tiffOpen: fileName mode: mode.
    ^result = 0 ifFalse: [ ExternalHandle fromInteger: result ]!

open: fileName mode: mode onFile: aFile
    " Private - Open a TIFF on an opened file descriptor for read/writing.
    Note that TIFFFdOpen and TIFFOpen recognise the character 'u' in the mode
    string, which forces the file to be opened unmapped.
    Returns aTiffHandle with opened image or nil if failed."

    ^self
        open: fileName
        mode: mode
        onFileHandle: aFile fileId!

open: fileName mode: mode onFileHandle: aFileHandle
    " Private - Open a TIFF on an opened file descriptor for read/writing.
    Note that TIFFFdOpen and TIFFOpen recognise the character 'u' in the mode
    string, which forces the file to be opened unmapped.
    Returns aTiffHandle with opened image or nil if failed."

    | result |
    result := self
        fileDescriptor: aFileHandle asParameter
        open: fileName
        mode: mode.
    ^result = 0 ifFalse: [ ExternalHandle fromInteger: result ]!

open: fileName mode: mode onFileStream: aFileStream
    " Private - Open a TIFF on an opened file descriptor for read/writing.
    Note that TIFFFdOpen and TIFFOpen recognise the character 'u' in the mode
    string, which forces the file to be opened unmapped.
    Returns aTiffHandle with opened image or nil if failed."

    ^self
        open: fileName
        mode: mode
        onFile: aFileStream file!
  
osCheckcmap: count red: red green: green blue: blue
    " Private - Utility function for DIB conversion.
    int osCheckcmap(int n, uint16* r, uint16* g, uint16* b)
    {
    while (n-- > 0)
        if (*r++ >= 256 || *g++ >= 256 || *b++ >= 256)
        return (16);
    return (8);
    }
    "

    <api: osCheckcmap long struct struct struct long>
    ^self invalidArgument!

predictorInit: tifHandle
    "Predictor Tag Support initialization."

    <api: TIFFPredictorInit handle long>
    ^self invalidArgument!
   
printDirectory: tifHandle file: file flags: flags
    "Print the contents of the current directory to the specified stdio file stream."

    <api: TIFFPrintDirectory handle handle long none>
    ^self invalidArgument!
   
rasterScanlineSize: tifHandle
    "Return the number of bytes required to store a complete decoded and packed raster
     scanline (as opposed to the I/O size returned by TIFFScanlineSize which may be less
     if data is store as separate planes)."

    <api: TIFFRasterScanlineSize handle long>
    ^self invalidArgument!
   
readBufferSetup: tifHandle into: lpData size: size
    "Setup the raw data buffer in preparation for reading a strip of raw data.
    If the buffer is specified as zero, then a buffer of appropriate size is allocated by the library.
    Otherwise, the client must guarantee that the buffer is large enough to hold any individual strip of raw data."

    <api: TIFFReadBufferSetup handle struct long long>
    ^self invalidArgument!
   
readDirectory: tifHandle
    "Read the next TIFF directory from a file and convert it to the internal format.
    We read directories sequentially."

    <api: TIFFReadDirectory handle boolean>
    ^self invalidArgument!
   
readEncodedStrip: tifHandle number: number into: buffer size: size
    "Read a strip of data and decompress the specified amount into the user-supplied buffer."

    <api: TIFFReadEncodedStrip handle ulong struct long long>
    ^self invalidArgument!
  
readEncodedTile: tifHandle number: number into: buffer size: size
    "Read a tile of data and decompress the specified amount into the user-supplied buffer."

    <api: TIFFReadEncodedTile handle ulong struct long long>
    ^self invalidArgument!
 
readRawStrip: tifHandle number: number into: buffer size: size
    "Read a strip of data from file."

    <api: TIFFReadRawStrip handle ulong struct long long>
    ^self invalidArgument!
  
readRawTile: tifHandle number: number into: buffer size: size
    "Read a tile of data from file."

    <api: TIFFReadRawTile handle ulong struct long long>
    ^self invalidArgument!
 
readRgbaImage: tiffHandle width: width height: height raster: lpRaster stop: stop
    "Read the specified image into an ABGR-format raster.
    This call is implemented as:
/* Read the specified image into an ABGR-format raster. */
int
TIFFReadRGBAImage(TIFF* tif,
    uint32 rwidth, uint32 rheight, uint32* raster, int stop)
{
    char emsg[1024];
    TIFFRGBAImage img;
    int ok;

    if (TIFFRGBAImageBegin(&img, tif, stop, emsg)) {
    /* XXX verify rwidth and rheight against width and height */
    ok = TIFFRGBAImageGet(&img, raster+(rheight-img.height)*rwidth,
        rwidth, img.height);
    TIFFRGBAImageEnd(&img);
    } else {
    TIFFError(TIFFFileName(tif), emsg);
    ok = 0;
    }
    return (ok);
}
    "

    <api: TIFFReadRGBAImage handle ulong ulong struct long boolean>
    ^self invalidArgument!
   
readScanLine: tifHandle into: lpData row: row mode: mode
    "Read a scanline of data."

    <api: TIFFReadScanline handle struct ulong ushort long>
    ^self invalidArgument!
 
readTile: tifHandle into: lpData  x: x y: y z: z sample: sample
    "Read and decompress a tile of data.
    The tile is selected by the (x,y,z,s) coordinates."

    <api: TIFFReadTile handle struct ulong ulong ulong ushort long>
    ^self invalidArgument!
   
registerCodec: scheme name: aString initMethod: initMethod
    " Private - Register and returns the codec pointer for a new scheme codec."

    <api: TIFFRegisterCODEC ushort struct ulong ulongReturn>
    ^self invalidArgument!
 
reverseBits: lpBuffer count: count
    "Load lpBuffer with BitRevTable bits."

    <api: TIFFReverseBits struct ulong none>
    ^self invalidArgument!
  
rgbaImageBegin: rgbaImageStruc from: tiffHandle stop: stop message: aString
    "Mark RGBA image scanning start process."

    <api: TIFFRGBAImageBegin ulong handle long struct boolean>
    ^self invalidArgument!

rgbaImageEnd: rgbaImageStruc
    "Mark RGBA image scanning end process."

    <api: TIFFRGBAImageEnd ulong none>
    ^self invalidArgument!
 
rgbaImageGet: rgbaImageStruc raster: lpRaster width: width height: height
    "Get RGBA image data."

    <api: TIFFRGBAImageGet ulong struct ulong ulong boolean>
    ^self invalidArgument!
   
rgbaImageOk: rgbaHandle message: aString
    "Check the image to see if TIFFReadRGBAImage can deal with it.
     aBoolean is returned according to whether or not the image can be handled.
    If false is returned, aString contains the reason why it is being rejected."

    <api: TIFFRGBAImageOK handle struct boolean>
    ^self invalidArgument!
 
scanlineSize: tifHandle
    "Return the number of bytes to read/write in a call to one of the scanline-oriented i/o routines.
    Note that this number may be 1 or samples-per-pixel if data is stored as separate planes."

    <api: TIFFScanlineSize handle long>
    ^self invalidArgument!
   
setCodec: scheme named: aString sending: initSelector to: aReceiver
    "Register and returns a new scheme codec (or nil if failed)."

    | callBack result |
    callBack := self class
        newCallBack: #codecInitMethod
        sending: initSelector
        to: aReceiver.
    result := self codecAtAddress: (
        self
            registerCodec: scheme
            name: aString asParameter
            initMethod: callBack).
    result isNil ifTrue: [ callBack release ].
    ^result!
 
setCompressionScheme: tifHandle scheme: scheme
    "Set the compression scheme for a tiff image."

    <api: TIFFSetCompressionScheme handle long long>
    ^self invalidArgument!
  
setDirectory: tifHandle to: aNumber
    "Set the n-th directory as the current directory.
    Directories are numbered starting at 0."

    <api: TIFFSetDirectory handle ushort boolean>
    ^self invalidArgument!
   
setErrorHandler: aHandler
    "Set the error handler routine.
    Returns the last (replaced) handler."

    <api: TIFFSetErrorHandler ulong ulongReturn>
    ^self invalidArgument!
   
setField: tiffHandle tag: tag value: value
    "Set a tag's value in the current directory."

    <api: TIFFSetField handle long struct boolean>
    ^self invalidArgument!
 
setField: tiffHandle tag: tag with: value1 with: value2 with: value3
    " Private - Sets a tag's value in the current tiff directory. "

    <api: TIFFSetField handle long struct struct struct boolean>
    ^self invalidArgument!
   
setIntegerField: tiffHandle tag: tag value: value
    "Set a tag's value in the current directory."

    <api: TIFFSetField handle long long boolean>
    ^self invalidArgument!

setSubDirectory: tiffHandle offset: offset
    "Set the current directory to be the directory located at the specified file offset.
    This interface is used mainly to access directories linked with the SubIFD tag
    (e.g. thumbnail images)."

    <api: TIFFSetSubDirectory handle ulong boolean>
    ^self invalidArgument!
  
setTagExtender: extender
    "Set Client Tag extension support callback."

    <api: TIFFSetTagExtender ulong ulongReturn>
    ^self invalidArgument!
   
setWarningHandler: tiffErrorHandler
    "Set Client Warning support callback."

    <api: TIFFSetWarningHandler ulong ulongReturn>
    ^self invalidArgument!
   
setWriteOffset: tifHandle to: anOffset
    "Set the current write offset.
    This should only be used to set the offset to a known previous location
    (very carefully), or to 0 so that the next write gets appended to the end of the file."

    <api: TIFFSetWriteOffset handle long none>
    ^self invalidArgument!
  
stripSize: tifHandle
    "Compute the # bytes in a (row-aligned) strip.
    Note that if RowsPerStrip is larger than the recorded ImageLength,
     then the strip size is truncated to reflect the actual space required
     to hold the strip."

    <api: TIFFStripSize handle long>
    ^self invalidArgument!
  
swabArrayOfDouble: lpDoubleArray count: count
    "TIFF Library Bit & Byte Swapping Support.
    Swap thebytes of double numbers in the holder Array"

    <api: TIFFSwabArrayOfDouble ulong ulong none>
    ^self invalidArgument!

swabArrayOfLong: lpLongArray count: count
    "TIFF Library Bit & Byte Swapping Support.
    Swap the bytes of a 32Bit integers in the holder Array"

    <api: TIFFSwabArrayOfLong ulong ulong none>
    ^self invalidArgument!
   
swabArrayOfShort: lpShortArray count: count
    "TIFF Library Bit & Byte Swapping Support.
    Swap the low and hight bytes of 16Bit integers in the holder Array"

    <api: TIFFSwabArrayOfShort ulong ulong none>
    ^self invalidArgument!

swabDouble: lpDouble
    "TIFF Library Bit & Byte Swapping Support.
    Swap thebytes of a double structure"

    <api: TIFFSwabDouble struct none>
    ^self invalidArgument!
 
swabLong: lpLong
    "TIFF Library Bit & Byte Swapping Support.
    Swap the bytes of a 32Bit integer"

    <api: TIFFSwabLong struct none>
    ^self invalidArgument!
 
swabShort: lpShort
    "TIFF Library Bit & Byte Swapping Support.
    Swap the low and hight bytes of a 16Bit integer."

    <api: TIFFSwabShort struct none>
    ^self invalidArgument!
   
tiffClientOpen: fileName mode: mode
    clientdata: handle
    readproc: readproc writeproc: writeproc seekproc: seekproc closeproc: closeproc
    sizeproc: sizeproc mapproc: mapproc unmapproc: unmapproc
    "Open a file for reading or writing.
    Returns a Tiff handle or zero if failed."

    <api: TIFFClientOpen
        struct struct ulong
        ulong ulong ulong ulong
        ulong ulong ulong
        long>
    ^self invalidArgument!
 
tiffGetBitRevTable: reversed
    " Private - Bit reversal tables.
    TIFFBitRevTable[<byte>] gives the bit reversed value of <byte>.
    Used in various places in the library when the FillOrder requires bit reversal of byte values (e.g. CCITT Fax 3 encoding/decoding).
    TIFFNoBitRevTable is provided for algorithms that want an equivalent table that do not reverse bit values.
    Returns a pointer to TIFFBitRevTable array."

    <api: TIFFGetBitRevTable boolean ulongReturn>
    ^self invalidArgument!
 
tiffOpen: fileName mode: mode
    " Private - Open a file for reading or writing.
    Returns a Tiff address or zero if failed."

    <api: TIFFOpen struct struct long>
    ^self invalidArgument!

tileRowSize: tifHandle
    "Compute the # bytes in each row of a tile."

    <api: TIFFTileRowSize handle long>
    ^self invalidArgument!
  
tileSize: tifHandle
    "Compute the # bytes in a row-aligned tile."

    <api: TIFFTileSize handle long>
    ^self invalidArgument!

unlinkDirectory: tifHandle number: aNumber
    "Unlink the specified directory from the directory chain."

    <api: TIFFUnlinkDirectory handle ushort boolean>
    ^self invalidArgument!
  
unRegisterCodec: aCodec
    " Unregister the codec structure aCodec."

    <api: TIFFUnRegisterCODEC struct none>
    ^self invalidArgument!

version
    "Get the version string."

    | address |
    address := ExternalGlobalAddress fromInteger: self getVersion.
    ^String fromAddress: address!

vGetField: tiffHandle tag: tag list: vaList
    "Like TIFFGetField, but taking a varargs parameter list.
    This routine is useful for building higher-level interfaces on top of the library."

    <api: TIFFVGetField handle long struct long>
    ^self invalidArgument!
  
vGetFieldDefaulted: tiffHandle tag: tag list: vaList
    "Like TIFFGetField, but return any default value if the tag is not present in the directory.
    We use the value in the directory, rather than explcit values so that defaults exist only
     one place in the library -- in TIFFDefaultDirectory."

    <api: TIFFVGetFieldDefaulted handle long struct long>
    ^self invalidArgument!
  
vSetField: tiffHandle tag: tag list: vaList
    "Like TIFFSetField, but taking a varargs parameter list.
    This routine is useful for building higher-level interfaces on top of the library."

    <api: TIFFVSetField handle long struct boolean>
    ^self invalidArgument!
   
vStripSize: tiffHandle rows: rows
    "Compute the # bytes in a variable height, row-aligned strip."

    <api: TIFFVStripSize handle ulong long>
    ^self invalidArgument!

vTileSize: tiffHandle rows: rows
    "Compute the # bytes in a variable length, row-aligned tile."

    <api: TIFFVTileSize handle ulong long>
    ^self invalidArgument!
   
writeBufferSetup: tiffHandle into: lpBuffer size: size
    "Setup the raw data buffer used for encoding."

    <api: TIFFWriteBufferSetup handle struct long long>
    ^self invalidArgument!
   
writeDirectory: tiffHandle
    "Write the contents of the current directory to the specified file.
    This routine doesn't handle overwriting a directory with auxiliary storage that's been changed."

    <api: TIFFWriteDirectory handle boolean>
    ^self invalidArgument!
   
writeEncodedStrip: tiffHandle strip: aNumber into: lpBuffer size: size
    "Encode the supplied data and write it to the specified strip.
    There must be space for the data; we don't check if strips overlap.
    Image length must be setup before writing."

    <api: TIFFWriteEncodedStrip handle ulong struct long long>
    ^self invalidArgument!
  
writeEncodedTile: tiffHandle tile: aNumber into: lpBuffer size: size
    "Encode the supplied data and write it to the specified tile.
    There must be space for the data.
    The function clamps individual writes to a tile to the tile size,
     but does not (and can not) check that multiple writes to the
     same tile do not write more than tile size data.
    Image length must be setup before writing; this interface
     does not support automatically growing the image on each write
     (as TIFFWriteScanline does)."

    <api: TIFFWriteEncodedTile handle ulong struct long long>
    ^self invalidArgument!

writeRawStrip: tiffHandle strip: aNumber into: lpBuffer size: size
    "Write the supplied data to the specified strip.
    There must be space for the data; we don't check if strips overlap.
    Image length must be setup before writing."

    <api: TIFFWriteRawStrip handle ulong struct long long>
    ^self invalidArgument!

writeRawTile: tiffHandle tile: aNumber into: lpBuffer size: size
    "Write the supplied data to the specified strip.
    There must be space for the data; we don't check if strips overlap.
    Image length must be setup before writing; this interface
     does not support automatically growing the image
     on each write (as TIFFWriteScanline does)."

    <api: TIFFWriteRawTile handle ulong struct long long>
    ^self invalidArgument!

writeScanLine: tifHandle into: lpBuffer number: scanNumber mode: mode
    "Write a scanline of data"

    <api: TIFFWriteScanline handle struct ulong ushort long>
    ^self invalidArgument!
   
writeTile: tifHandle into: lpBuffer x: x y: y z: z sample: sample
    "Write and compress a tile of data.
    The tile is selected by the (x,y,z,s) coordinates."

    <api: TIFFWriteTile handle struct ulong ulong ulong ushort long>
    ^self invalidArgument! !



!TIFFExample methods !
 
dibInspect
    "Let's the user select a file and inspect the receiver as aDIB. "

    | dib |
    self tiffFromUser isNil ifTrue: [^nil].
    dib := tiff asDIB.
    self release.
    ^dib inspect!
 
dirCount
    "Count file directories of a tiff file selected from user. 
	TIFFExample new dirCount.
	"

    | count pathName |
    self tiffFromUser isNil ifTrue: [^nil].
    count := 0.
    tiff withDirectoriesDo: [ count := count + 1 ].
	pathName := tiff pathName.
    self release.
    MessageBox message: 'The image in file ', pathName asString
        ,' has ',(count > 0 ifTrue: [count asString] ifFalse: ['No']),' directory entries.'!
   
initializeNotifier
    " Private - Initializes the window for receiver's notification to user. "

    notifier := TextWindow
        windowLabeled: self class name,' notifier',
            (tiff isNil ifTrue: ['.'] ifFalse: [' on ',tiff pathName asString])
        frame: (0@0 extent: Display extent // 2)!

notifier
    " Private - Returns an open text window for receiver's notification to user. "

    notifier isNil ifTrue: [ self initializeNotifier ].
    ^notifier!
 
release
    "Release the receiver."

    tiff release.
    tiff := nil.

    notifier notNil ifTrue: [
        self notifier cr;show: 'Done...';cr
    ].

    ^super release!

showContents
    "Show the contents of a file selected from user.
	TIFFExample new showContents.
	"

    self tiffFromUser isNil ifTrue: [^nil].
    self notifier pane font: Font fixedSystemFont.
    self notifier
        cr;show: 'Showing TIFF contents bytes.'.
    tiff withContentsDo: [:buffer|
        self notifier cr.
        buffer dumpOn: self notifier base: 16
    ].
    self release.!

showRGBA
    "Show the RGBA bytes of a file selected from user.
	TIFFExample new showRGBA.
	"

    | rgba |
    self tiffFromUser isNil ifTrue: [^nil].
    self notifier pane font: Font fixedSystemFont.
    self notifier
        cr;show: 'Showing RGBA contents bytes.'.
    rgba := tiff getRGBA.
    self release.
    rgba dumpOn: self notifier base: 16.!
 
showScanLines
    "Show the ScanLines of a file selected from user.
	TIFFExample new showScanLines.
	"

    | line |
    self tiffFromUser isNil ifTrue: [^nil].
    self notifier pane font: Font fixedSystemFont.
    line := 1.
    self notifier
        cr;show: 'Showing TIFF ScanLines bytes. (scanlineSize='
            ,tiff scanlineSize asString,').'.
    tiff withScanLinesDo: [:buffer :plane |
        self notifier cr;show:  ' Line: ',line asString,' Plane: ',plane asString;cr.
        buffer dumpOn: self notifier base: 16.
        line := line + 1.
    ].
    self release.!

showStripLines
    "Show the Strips of a file selected from user.
	TIFFExample new showStripLines.
	"

    | line |
    self tiffFromUser isNil ifTrue: [^nil].
    self notifier pane font: Font fixedSystemFont.
    line := 1.
    self notifier
        cr;show: 'Showing TIFF Strips bytes. (StripsSize='
            ,tiff stripSize asString,').'.
    tiff withStripsDo: [:buffer |
        self notifier cr;show:  ' Line: ',line asString;cr.
        buffer dumpOn: self notifier base: 16.
        line := line + 1.
    ].
    self release.!
 
showTiles
    "Show the Tiles of a file selected from user.
	TIFFExample new showTiles.
	"

    | line |
    self tiffFromUser isNil ifTrue: [^nil].
    tiff hasTiles
    ifTrue: [
        self notifier pane font: Font fixedSystemFont.
        line := 1.
        self notifier
            cr;show: 'Showing TIFF Tiles. (TileSize='
                ,tiff tileSize asString,').'.
        tiff withTilesDo: [:buffer |
            self notifier cr;show:  ' Tile: ',line asString;cr.
            buffer dumpOn: self notifier base: 16.
            line := line + 1.
        ].
    ]
    ifFalse: [
        MessageBox message: 'Image has no tiles.'
    ].
    self release.!
   
tiffFromUser
    " Private - Let's the user select a file and returns a TIFFImage instance from it (or nil). "

    | dialog file |
    dialog := FileDialog new.
    dialog
        title: 'Coose a tiff file to open';
        fileSpec: '*.tif';
        addFilter: '*.*' description: 'All Files (*.*)';
        addFilter: '*.tif' description: 'Tif Files (*.TIF)';
        addFilter: '*.tiff' description: 'Tiff Files (*.TIFF)';
        defFilter: '*.tif';
        hideReadonly;
        open.
    (file := dialog file) isNil ifTrue: [^nil].
    tiff := TIFFImage fromFile: file.
    ^tiff! !



!TIFFMemoryReader class methods !
   
on: aStream
    "Returns an instance of teh receiver for reading/writing on aStream."

    ^self new
        stream: aStream;
        yourself!

tiffLibrary
    " Private - Returns the TIFF library."

    ^TIFFDLL32 current! !



!TIFFMemoryReader methods !
 
closeCallback
    "Returns the instance var of the receiver
     for external manipulation"

    closeCallback isNil ifTrue: [ self initializeCloseCallback ].
    ^closeCallback!
 
closeCallback: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    closeCallback := anObject!
 
closeProc: handle
    " Private - Process the closeCallback.
    By default aStream will NOT be closed."

    ^0 "success -1=fail"!
 
imageSize
    "Returns the instance var of the receiver
     for external manipulation"
    ^imageSize!
  
imageSize: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    imageSize := anObject!
 
initializeCloseCallback
    " Private - Initializes the receiver's closeCallback"

    closeCallback := self newCallBack: #closeProc sending: #closeProc:.!
  
initializeMapCallback
    " Private - Initializes the receiver's mapCallback"

    mapCallback := self newCallBack: #mapFileProc sending: #mapProc:with:with:.!
  
initializeReadCallback
    " Private - Initializes the receiver's readCallback"

    readCallback := self newCallBack: #readWriteProc sending: #readProc:with:with:.!

initializeSeekCallback
    " Private - Initializes the receiver's seekCallback"

    seekCallback := self newCallBack: #seekProc sending: #seekProc:with:with:.!
 
initializeSizeCallback
    " Private - Initializes the receiver's sizeCallback"

    sizeCallback := self newCallBack: #sizeProc sending: #sizeProc:.!
   
initializeUnmapCallback
    " Private - Initializes the receiver's unmapCallback"

    unmapCallback := self newCallBack: #unmapFileProc sending: #unmapProc:with:with:.!

initializeWriteCallback
    " Private - Initializes the receiver's writeCallback"

    writeCallback := self newCallBack: #readWriteProc sending: #writeProc:with:with:.!

mapCallback
    "Returns the instance var of the receiver
     for external manipulation"

    mapCallback isNil ifTrue: [ self initializeMapCallback ].
    ^mapCallback!
 
mapCallback: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    mapCallback := anObject!
 
mapProc: lpOfs with: lpData with: handle
    " Private - Process the mapCallback."

    "
	| data ofs |
    data := ExternalAddress fromInteger: lpData.
    ofs := ExternalAddress fromInteger: lpOfs.
	"
    ^0!
  
newCallBack: callBackName sending: selector
    " Private - Returns a valid callback to process with selector."

    ^self tiffLibrary class
        newCallBack: callBackName
        sending: selector
        to: self!

open: pathName
    "Returns a handle to a Tiff image opened for reading using the receiver."

    ^self tiffLibrary
        clientOpen: pathName asParameter
        mode: 'r'
        clientdata: self hash asParameter
        readproc: self readCallback asParameter
        writeproc: self writeCallback asParameter
        seekproc: self seekCallback asParameter
        closeproc: self closeCallback asParameter
        sizeproc: self sizeCallback asParameter
        mapproc: self mapCallback asParameter
        unmapproc: self unmapCallback asParameter.!
 
readCallback
    "Returns the instance var of the receiver
     for external manipulation"

    readCallback isNil ifTrue: [ self initializeReadCallback ].
    ^readCallback!
 
readCallback: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    readCallback := anObject!
   
readProc: count with: lpBuffer with: handle
    " Private - Process the readCallback."

    | anAddress collection |
    anAddress := ExternalAddress fromInteger: lpBuffer.
	collection := self stream next: count.
	anAddress replaceBytes: count with: collection.
    ^collection size!
  
release
	" Private - Release the receiver's resources."

	super release.
	self releaseCallbacks.!
   
releaseCallbacks
	" Private - Release the receiver's callbacks."

	readCallback release.
	writeCallback release.
	seekCallback release.
	closeCallback release.
	sizeCallback release.
	mapCallback release.
	unmapCallback release.

	readCallback := writeCallback := seekCallback := nil.
	closeCallback := sizeCallback := mapCallback := nil.
	unmapCallback := nil.!
 
seekCallback
    "Returns the instance var of the receiver
     for external manipulation"

    seekCallback isNil ifTrue: [ self initializeSeekCallback ].
    ^seekCallback!
 
seekCallback: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    seekCallback := anObject!
   
seekProc: method with: offset with: handle
    " Private - Process the seekCallback."

    | aStream |
    (aStream := self stream) isNil ifTrue: [ ^0 ].
    method = 2 ifTrue: [ aStream position: aStream size + offset ]
    ifFalse: [
        method = 1
        ifTrue: [ aStream position: aStream position + offset ]
        ifFalse: [ aStream position: offset ]
    ].
    ^aStream position!
   
sizeCallback
    "Returns the instance var of the receiver
     for external manipulation"

    sizeCallback isNil ifTrue: [ self initializeSizeCallback ].
    ^sizeCallback!
 
sizeCallback: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    sizeCallback := anObject!
   
sizeProc: handle
    " Private - Process the sizeCallback.
    By default returns imageSize or aStream size."

    ^imageSize notNil
        ifTrue: [ imageSize ]
        ifFalse: [ stream size ]!
  
stream
    "Returns the instance var of the receiver
     for external manipulation"
    ^stream!

stream: aStream
    "Sets the instance var of the receiver and compute image size from aStream size.
    Consider that the image ocuppies up to end of aStream"

    stream := aStream.
    self imageSize: aStream size - aStream position.!
  
tiffLibrary
    " Private - Returns the TIFF library."

    ^self class tiffLibrary!
 
unmapCallback
    "Returns the instance var of the receiver
     for external manipulation"

    unmapCallback isNil ifTrue: [ self initializeUnmapCallback ].
    ^unmapCallback!
 
unmapCallback: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    unmapCallback := anObject!
 
unmapProc: handle with: data with: ofs
    " Private - Process the unmapCallback."

    ^0!
  
writeCallback
    "Returns the instance var of the receiver
     for external manipulation"

    writeCallback isNil ifTrue: [ self initializeWriteCallback ].
    ^writeCallback!
 
writeCallback: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    writeCallback := anObject!
 
writeProc: count with: lpBuffer with: handle
    " Private - Process the writeCallback."

    | anAddress collection |
    collection := String new: count.
    collection replace: count with: (ExternalAddress fromInteger: lpBuffer).
    self stream nextPutAll: collection.
    ^collection size! !
